'''
See COPYRIGHT.md for copyright information.

This plugin allows GUI and command line users to test transforms.

Custom extensions are also made available when their plugin has been loaded
(e.g., for SEC custom transforms use plugin EDGAR/transforms, EDGAR/validate or EDGAR/render)

Errors are shown both in results field as well as reported in the message pane (GUI) and reported in log (command line).

For GUI operation tools -> transformation tester:

    Select registry (ixt v1-4, ixt-sec)
    Select or enter transform
    Enter source text
    Press "transform"
    Result (or error code)

For command line operation:

    arelleCmdLine --plugins transforms/tester --testTransform 'registry name transformation name pattern' (space separated)
    note: the transform name may be optionally prefixed
    results or errors are in the log

    arelleCmdLine --plugins transforms/tester --testTransform 'ixt v3 datedaymonthen 29th February' or
    arelleCmdLine --plugins transforms/tester --testTransform 'ixt v3 ixt:datedaymonthen 29th February'

    SEC transform example:

    arelleCmdLine --plugins 'transforms/tester|EDGAR/SEC' --testTransform "ixt-sec durwordsen 23 days"

    Help instructions and list of available transforms:

    arelleCmdLine --plugins transforms/tester --testTransform 'help'  (or '?') or
    arelleCmdLine --plugins 'transforms/tester|EDGAR/transforms' --testTransform 'help'

For REST API operation:
    web browser: http://localhost:8080/rest/xbrl/validation?plugins=transforms/tester&testTransform=ixt v3 datedaymonthen 29th February
    cmd line: curl 'http://localhost:8080/rest/xbrl/validation?plugins=transforms/tester&testTransform=ixt%20v3%20datedaymonthen%2029th%20February'

'''
import os, logging
import regex as re
from optparse import SUPPRESS_HELP
from arelle.FunctionIxt import ixtNamespaces, ixtNamespaceFunctions
from arelle.ModelFormulaObject import Trace
from arelle.Version import authorLabel, copyrightLabel
from arelle.XmlUtil import setXmlns
from arelle import ModelXbrl, ValidateXbrl, XbrlConst
from arelle.ModelDocumentType import ModelDocumentType
from arelle.formula import XPathContext, XPathParser

class TransformTester:
    def __init__(self, cntlr, isCmdLine=False):
        self.cntlr = cntlr

        # setup tester
        xml = "<rootElement/>"
        self.modelXbrl = ModelXbrl.create(cntlr.modelManager, ModelDocumentType.UnknownNonXML, initialXml=xml, isEntry=True)
        self.validator = ValidateXbrl.ValidateXbrl(self.modelXbrl)
        self.validator.validate(self.modelXbrl)  # required to set up
        cntlr.showStatus(_("Initializing Formula Grammar"))
        XPathParser.initializeParser(cntlr.modelManager)
        cntlr.showStatus(None)

        self.trRegs = sorted(ixtNamespaces.keys())
        self.trPrefixNSs = dict((qn.prefix, qn.namespaceURI)
                                  for qn in self.modelXbrl.modelManager.customTransforms.keys())
        self.trRegs.extend(sorted(self.trPrefixNSs.keys()))
        self.trPrefixNSs.update(ixtNamespaces)

    def getTrNames(self, trReg):
        trNS = self.trPrefixNSs[trReg]
        trPrefix = trReg.split()[0] # for ixt remove TRn part
        setXmlns(self.modelXbrl.modelDocument, trPrefix, trNS)
        if trNS in ixtNamespaceFunctions:
            return sorted("{}:{}".format(trPrefix, key)
                          for key in ixtNamespaceFunctions[trNS].keys())
        # custom transforms
        return sorted(str(trQn)
                      for trQn in self.modelXbrl.modelManager.customTransforms.keys()
                      if trQn.prefix == trPrefix)

    def transform(self, trReg, trName, sourceValue):
        try:
            trNS = self.trPrefixNSs[trReg]
            trPrefix = trReg.split()[0] # for ixt remove TRn part
            setXmlns(self.modelXbrl.modelDocument, trPrefix, trNS)
            self.modelXbrl.modelManager.showStatus(_("Executing call"))
            elt = self.modelXbrl.modelDocument.xmlRootElement
            if ':' in trName:
                prefixedFnName = trName
            else:
                prefixedFnName = "{}:{}".format(trPrefix, trName)
            callExprStack = XPathParser.parse(self.validator,
                                              '{}("{}")'.format(prefixedFnName, sourceValue),
                                              elt, trName + " call", Trace.CALL)
            xpathContext = XPathContext.create(self.modelXbrl, sourceElement=elt)
            result = xpathContext.evaluate(callExprStack)
            while result and isinstance(result, (tuple,list,set)):
                result = next(iter(result)) # de-sequence result
            return result
        except XPathContext.XPathException as err:
            self.modelXbrl.error(err.code, err.message)
            return err



def cmdLineOptionExtender(parser, *args, **kwargs):
    parser.add_option("--testTransform",
                      action="store",
                      dest="testTransform",
                      help=_("Test a transformation registry transform. "
                             "Enter 'help' or '?' for a list of transformation registries available.  "
                             "Enter registry name, space, transformation name, space and pattern.  "
                             "E.g., 'ixt v3 datedaymonthen 29th February' or ixt v3 ixt:datedaymonthen 29th February'. "))


def cmdLineRun(cntlr, options, *args, **kwargs):
    if options.testTransform:
        tester = TransformTester(cntlr)
        arg = options.testTransform
        argWord, _sep, rest = arg.partition(" ")
        trReg = None
        for _regName in tester.trPrefixNSs.keys():
            if arg.startswith(_regName):
                trReg = _regName
                rest = arg[len(trReg)+1:]
        if trReg is None or argWord in ("help", "?"):
            cntlr.addToLog("Registries available: {}".format(", ".join(tester.trRegs)),
                           messageCode="tester:registries", level=logging.INFO)
        else:
            trName, _sep, sourceValue = rest.partition(" ")
            print("reg {} name {} source {}".format(trReg, trName, sourceValue))
            result = tester.transform(trReg, trName, sourceValue)
            tester.modelXbrl.info("tester:transformation",
                                  "%(registry)s %(transformName)s source '%(sourceValue)s' result '%(resultValue)s'",
                                  registry=trReg, transformName=trName, sourceValue=sourceValue, resultValue=str(result))
        tester.modelXbrl.close()

def transformationTesterMenuExtender(cntlr, menu, *args, **kwargs):
    # define tkinger only when running in GUI mode, not available in cmd line or web server modes
    from tkinter import Toplevel, StringVar, N, S, E, EW, W
    try:
        from tkinter.ttk import Frame, Button, Entry
    except ImportError:
        from ttk import Frame, Button
    from arelle.UiUtil import gridHdr, gridCell, gridCombobox, label, checkbox
    from arelle.CntlrWinTooltip import ToolTip
    class DialogTransformTester(Toplevel):
        def __init__(self, tester):
            self.tester = tester
            self.mainWin = tester.cntlr

            parent = self.mainWin.parent
            super(DialogTransformTester, self).__init__(parent)
            self.parent = parent
            parentGeometry = re.match(r"(\d+)x(\d+)[+]?([-]?\d+)[+]?([-]?\d+)", parent.geometry())
            dialogX = int(parentGeometry.group(3))
            dialogY = int(parentGeometry.group(4))
            self.selectedGroup = None

            self.transient(self.parent)
            self.title(_("Transformation Tester"))

            frame = Frame(self)

            # load grid
            trRegLabel = label(frame, 0, 0, _("Registry:"))
            trReg = self.tester.trRegs[-1] # default is latest
            self.trRegName = gridCombobox(frame, 1, 0,
                                          value=trReg,
                                          values=self.tester.trRegs,
                                          comboboxselected=self.dialogTrRegComboBoxSelected)
            trRegToolTipMessage = _("Select Transformation Registry")
            ToolTip(self.trRegName, text=trRegToolTipMessage, wraplength=360)
            ToolTip(trRegLabel, text=trRegToolTipMessage, wraplength=360)

            trNameLabel = label(frame, 0, 1, _("Transform:"))
            self.trNameName = gridCombobox(frame, 1, 1,
                                          value="",
                                          values=self.tester.getTrNames(trReg),
                                          comboboxselected=self.dialogTrNameComboBoxSelected)
            trRegToolTipMessage = _("Select or enter transform")
            ToolTip(self.trRegName, text=trRegToolTipMessage, wraplength=360)
            ToolTip(trRegLabel, text=trRegToolTipMessage, wraplength=360)

            sourceLabel = label(frame, 0, 2, _("Source text:"))
            ToolTip(sourceLabel, text=_("Enter the source text which is to be transformed. "), wraplength=240)
            self.sourceVar = StringVar()
            self.sourceVar.set("")
            sourceEntry = Entry(frame, textvariable=self.sourceVar, width=50)
            sourceLabel.grid(row=2, column=0, sticky=W)
            sourceEntry.grid(row=2, column=1, sticky=EW, pady=3, padx=3)

            resultLabel = label(frame, 1, 3, _("Result:"))
            ToolTip(sourceLabel, text=_("Transformation result. "), wraplength=240)
            self.resultVar = StringVar()
            self.resultVar.set("")
            resultEntry = Entry(frame, textvariable=self.resultVar, width=50)
            resultLabel.grid(row=3, column=0, sticky=W)
            resultEntry.grid(row=3, column=1, sticky=EW, pady=3, padx=3)


            self.mainWin.showStatus(None)

            btnPad = 2 if self.mainWin.isMSW else 0 # buttons too narrow on windows
            okButton = Button(frame, text=_("Transform"), width=8 + btnPad, command=self.dialogOk)
            cancelButton = Button(frame, text=_("Done"), width=4 + btnPad, command=self.dialogClose)
            cancelButton.grid(row=4, column=0, sticky=E, columnspan=2, pady=3, padx=3)
            okButton.grid(row=4, column=0, sticky=E, columnspan=2, pady=3, padx=64)
            ToolTip(okButton, text=_("Transform the source entered. "), wraplength=240)
            ToolTip(cancelButton, text=_("Close this dialog. "), wraplength=240)

            frame.grid(row=0, column=0, sticky=(N,S,E,W))
            frame.columnconfigure(1, weight=3)
            frame.columnconfigure(2, weight=1)
            window = self.winfo_toplevel()
            window.columnconfigure(0, weight=1)
            self.geometry("+{0}+{1}".format(dialogX+150,dialogY+100))

            #self.bind("<Return>", self.ok)
            #self.bind("<Escape>", self.close)

            self.protocol("WM_DELETE_WINDOW", self.dialogClose)
            self.grab_set()

            self.wait_window(self)

        def dialogOk(self, event=None):
            result = self.tester.transform(
                self.trRegName.get(),
                self.trNameName.value,
                self.sourceVar.get())
            if isinstance(result, XPathContext.XPathException):
                self.resultVar.set(str(result))
            else:
                self.resultVar.set(str(result))

        def dialogClose(self, event=None):
            self.tester.modelXbrl.close()
            self.parent.focus_set()
            self.destroy()

        def dialogTrRegComboBoxSelected(self, *args):
            self.trNameName["values"] = self.tester.getTrNames( self.trRegName.get() )

        def dialogTrNameComboBoxSelected(self, *args):
            pass

    def guiTransformationTester():
        tester = TransformTester(cntlr, isCmdLine=True)
        DialogTransformTester( tester )
        tester.modelXbrl.close()

    menu.add_command(label="Transformtion Tester",
                     underline=0,
                     command=lambda: guiTransformationTester() )


__pluginInfo__ = {
    # Do not use _( ) in pluginInfo itself (it is applied later, after loading
    'name': 'Transformation tester',
    'version': '1.0',
    'description': '''Transformation Tester''',
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'CntlrCmdLine.Options': cmdLineOptionExtender,
    'CntlrCmdLine.Utility.Run': cmdLineRun,
    'CntlrWinMain.Menu.Tools': transformationTesterMenuExtender

}



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from enum import Enum
from functools import cached_property, lru_cache
from pathlib import Path

from regex import Pattern

from . import Constants


class ReportFolderType(Enum):
    ATTACH_DOC = "AttachDoc"
    AUDIT_DOC = "AuditDoc"
    PRIVATE_ATTACH = "PrivateAttach"
    PRIVATE_DOC = "PrivateDoc"
    PUBLIC_DOC = "PublicDoc"

    @classmethod
    def parse(cls, value: str) -> ReportFolderType | None:
        try:
            return cls(value)
        except ValueError:
            return None

    @cached_property
    def extensionCategory(self) -> ExtensionCategory | None:
        return FORM_TYPE_EXTENSION_CATEGORIES.get(self, None)

    @cached_property
    def isAttachment(self) -> bool:
        return "Attach" in self.value

    @cached_property
    def ixbrlFilenamePatterns(self) -> list[Pattern[str]]:
        return IXBRL_FILENAME_PATTERNS.get(self, [])

    @cached_property
    def manifestName(self) -> str:
        return f'manifest_{self.value}.xml'

    @cached_property
    def manifestPath(self) -> Path:
        return self.xbrlDirectory / self.manifestName

    @cached_property
    def namespaceUriPatterns(self) -> list[Pattern[str]]:
        return NAMESPACE_URI_PATTERNS.get(self, [])

    @cached_property
    def prefixPatterns(self) -> list[Pattern[str]]:
        return PREFIX_PATTERNS.get(self, [])

    @cached_property
    def xbrlDirectory(self) -> Path:
        return Path('XBRL') / str(self.value)

    @cached_property
    def xbrlFilenamePatterns(self) -> list[Pattern[str]]:
        return XBRL_FILENAME_PATTERNS.get(self, [])

    @lru_cache(1)
    def getValidExtensions(self, isAmendment: bool, isSubdirectory: bool) -> frozenset[str] | None:
        if self.extensionCategory is None:
            return None
        return self.extensionCategory.getValidExtensions(isAmendment, isSubdirectory)


class ExtensionCategory(Enum):
    ATTACH = 'ATTACH'
    DOC = 'DOC'

    def getValidExtensions(self, isAmendment: bool, isSubdirectory: bool) -> frozenset[str] | None:
        amendmentMap = VALID_EXTENSIONS[isAmendment]
        categoryMap = amendmentMap.get(self, None)
        if categoryMap is None:
            return None
        return categoryMap.get(isSubdirectory, None)


FORM_TYPE_EXTENSION_CATEGORIES = {
    ReportFolderType.ATTACH_DOC: ExtensionCategory.ATTACH,
    ReportFolderType.AUDIT_DOC: ExtensionCategory.DOC,
    ReportFolderType.PRIVATE_ATTACH: ExtensionCategory.ATTACH,
    ReportFolderType.PRIVATE_DOC: ExtensionCategory.DOC,
    ReportFolderType.PUBLIC_DOC: ExtensionCategory.DOC,
}


HTML_EXTENSIONS = frozenset({'.htm', '.html', '.xhtml'})
IMAGE_EXTENSIONS = frozenset({'.jpeg', '.jpg', '.gif', '.png'})
ASSET_EXTENSIONS = frozenset(HTML_EXTENSIONS | IMAGE_EXTENSIONS)
XBRL_EXTENSIONS = frozenset(HTML_EXTENSIONS | {'.xml', '.xsd'})
ATTACH_EXTENSIONS = frozenset(HTML_EXTENSIONS | {'.pdf', })

# Is Amendment -> Category -> Is Subdirectory
VALID_EXTENSIONS = {
    False: {
        ExtensionCategory.ATTACH: {
            False: ATTACH_EXTENSIONS,
            True: ASSET_EXTENSIONS,
        },
        ExtensionCategory.DOC: {
            False: XBRL_EXTENSIONS,
            True: ASSET_EXTENSIONS
        },
    },
    True: {
        ExtensionCategory.ATTACH: {
            False: ATTACH_EXTENSIONS,
            True: ASSET_EXTENSIONS,
        },
        ExtensionCategory.DOC: {
            False: HTML_EXTENSIONS,
            True: ASSET_EXTENSIONS,
        },
    },
}

IXBRL_FILENAME_PATTERNS = {
    ReportFolderType.AUDIT_DOC: [
        Constants.AUDIT_IXBRL_FILENAME_PATTERN,
        Constants.AUDIT_LINKBASE_FILENAME_PATTERN,
        Constants.AUDIT_SCHEMA_FILENAME_PATTERN,
    ],
    ReportFolderType.PUBLIC_DOC: [
        Constants.REPORT_COVER_FILENAME_PATTERN,
        Constants.REPORT_IXBRL_FILENAME_PATTERN,
        Constants.REPORT_LINKBASE_FILENAME_PATTERN,
        Constants.REPORT_SCHEMA_FILENAME_PATTERN,
    ]
}

NAMESPACE_URI_PATTERNS = {
    ReportFolderType.AUDIT_DOC: [
        Constants.AUDIT_NAMESPACE_PATTERN,
    ],
    ReportFolderType.PUBLIC_DOC: [
        Constants.REPORT_NAMESPACE_PATTERN,
    ],
}

PREFIX_PATTERNS = {
    ReportFolderType.AUDIT_DOC: [
        Constants.AUDIT_PREFIX_PATTERN,
    ],
    ReportFolderType.PUBLIC_DOC: [
        Constants.REPORT_PREFIX_PATTERN,
    ],
}

XBRL_FILENAME_PATTERNS = {
    ReportFolderType.AUDIT_DOC: [
        Constants.AUDIT_XBRL_FILENAME_PATTERN,
    ],
    ReportFolderType.PUBLIC_DOC: [
        Constants.REPORT_XBRL_FILENAME_PATTERN,
    ],
}



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from collections import defaultdict
from typing import Any, Iterable

from arelle import XbrlConst
from arelle.ModelDocumentType import ModelDocumentType
from arelle.ModelDtsObject import ModelResource, ModelConcept
from arelle.ValidateXbrl import ValidateXbrl
from arelle.typing import TypeGetText
from arelle.utils.PluginHooks import ValidationHook
from arelle.utils.validate.Decorator import validation
from arelle.utils.validate.Validation import Validation
from ..DisclosureSystems import (DISCLOSURE_SYSTEM_EDINET)
from ..PluginValidationDataExtension import PluginValidationDataExtension

_: TypeGetText


@validation(
    hook=ValidationHook.XBRL_FINALLY,
    disclosureSystems=[DISCLOSURE_SYSTEM_EDINET],
)
def rule_frta_2_1_9(
        pluginData: PluginValidationDataExtension,
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
) -> Iterable[Validation]:
    """
    EDINET.EC5710W: [FRTA.2.1.9] All documentation of a concept must be contained
    in XBRL linkbases. Do not use the xsd:documentation element in an element definition.
    """
    errors = []
    for modelDocument in val.modelXbrl.urlDocs.values():
        if pluginData.isStandardTaxonomyUrl(modelDocument.uri, val.modelXbrl):
            continue
        rootElt = modelDocument.xmlRootElement
        for elt in rootElt.iterdescendants(XbrlConst.qnXsdDocumentation.clarkNotation):
            errors.append(elt)
    if len(errors) > 0:
        yield Validation.warning(
            codes='EDINET.EC5710W.FRTA.2.1.9',
            msg=_("All documentation of a concept must be contained in XBRL linkbases. "
                  "Taxonomy element declarations should not use the XML Schema documentation element."),
            modelObject=errors,
        )


@validation(
    hook=ValidationHook.XBRL_FINALLY,
    disclosureSystems=[DISCLOSURE_SYSTEM_EDINET],
)
def rule_frta_2_1_10(
        pluginData: PluginValidationDataExtension,
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
) -> Iterable[Validation]:
    """
    EDINET.EC5710W: [FRTA.2.1.10] All extension taxonomy concepts must have a standard label.
    """
    errors = []
    for concept in val.modelXbrl.qnameConcepts.values():
        if pluginData.isStandardTaxonomyUrl(concept.modelDocument.uri, val.modelXbrl):
            continue
        if not concept.label(XbrlConst.standardLabel, fallbackToQname=False):
            errors.append(concept)
    if len(errors) > 0:
        yield Validation.warning(
            codes='EDINET.EC5710W.FRTA.2.1.10',
            msg=_("All extension taxonomy concepts must have a standard label. "
                  "A standard label is not specified for a concept in an "
                  "extension taxonomy. When adding a concept to an extension taxonomy, "
                  "please provide Japanese and English labels in the standard, verbose, and "
                  "documentation roles."),
            modelObject=errors,
        )


@validation(
    hook=ValidationHook.XBRL_FINALLY,
    disclosureSystems=[DISCLOSURE_SYSTEM_EDINET],
)
def rule_frta_2_1_11(
        pluginData: PluginValidationDataExtension,
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
) -> Iterable[Validation]:
    """
    EDINET.EC5710W: [FRTA.2.1.11] All concepts within a taxonomy schema should have a
    unique label for the standard or verbose role in each language used.
    """
    labelsRelationshipSet = val.modelXbrl.relationshipSet(XbrlConst.conceptLabel)
    labelGroupsByLang: dict[str, dict[str, dict[str, set[ModelConcept]]]] = (
        defaultdict(lambda: defaultdict(lambda: defaultdict(set)))
    )
    relevantRoles = frozenset({
        XbrlConst.standardLabel,
        XbrlConst.verboseLabel
    })
    for concept, modelLabelRels in labelsRelationshipSet.fromModelObjects().items():
        if pluginData.isStandardTaxonomyUrl(concept.modelDocument.uri, val.modelXbrl):
            continue
        for modelLabelRel in modelLabelRels:
            modelLabel = modelLabelRel.toModelObject
            if not isinstance(modelLabel, ModelResource):
                continue
            if modelLabel.role not in relevantRoles:
                continue
            if not modelLabel.modelDocument.inDTS:
                continue
            if not modelLabel.xmlLang:
                continue
            labelGroupsByLang[modelLabel.xmlLang][modelLabel.role][modelLabel.textValue].add(concept)
    for lang, labelGroupsByRole in labelGroupsByLang.items():
        duplicateLabelsByRole = defaultdict(list)
        conceptsWithVerboseLabel = {
            concept: False
            for concept in val.modelXbrl.qnameConcepts.values()
            if not pluginData.isStandardTaxonomyUrl(concept.modelDocument.uri, val.modelXbrl)
        }
        for role, labelGroupsByValue in labelGroupsByRole.items():
            for value, concepts in labelGroupsByValue.items():
                if role == XbrlConst.verboseLabel:
                    for concept in concepts:
                        conceptsWithVerboseLabel[concept] = True
                if len(concepts) < 2:
                    continue
                duplicateLabelsByRole[role].append((value, concepts))
        if len(duplicateLabelsByRole[XbrlConst.standardLabel]) == 0:
            # There are no duplicate standard labels, so we don't need to check verbose labels
            continue
        conceptsWithoutVerboseLabel = {concept for concept, value in conceptsWithVerboseLabel.items() if not value}
        if len(duplicateLabelsByRole[XbrlConst.verboseLabel]) == 0 and len(conceptsWithoutVerboseLabel) == 0:
            # All concepts have a verbose label, and there are no duplicate verbose labels,
            return
        for role, duplicateLabels in duplicateLabelsByRole.items():
            for value, concepts in duplicateLabels:
                yield Validation.warning(
                    codes='EDINET.EC5710W.FRTA.2.1.11',
                    msg=_("All concepts within a taxonomy schema should have a unique label for the "
                          "standard or verbose role in each language used. "
                          "The %(role)s label contains a duplicate label ('%(value)s') in the same "
                          "language ('%(lang)s'). Define either the standard label or the verbose label "
                          "so that they are unique in the same language."),
                    role='standard' if role == XbrlConst.standardLabel else 'verbose',
                    value=value,
                    lang=lang,
                    modelObject=concepts,
                )


@validation(
    hook=ValidationHook.XBRL_FINALLY,
    disclosureSystems=[DISCLOSURE_SYSTEM_EDINET],
)
def rule_frta_3_1_10(
        pluginData: PluginValidationDataExtension,
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
) -> Iterable[Validation]:
    """
    EDINET.EC5710W: [FRTA.3.1.10] Role types defined in the extension taxonomy must have a definition.
    """
    errors = []
    for uri, roleTypes in val.modelXbrl.roleTypes.items():
        for roleType in roleTypes:
            if pluginData.isStandardTaxonomyUrl(roleType.document.uri, val.modelXbrl):
                continue
            if not roleType.definition:
                errors.append(roleType)
    if len(errors) > 0:
        yield Validation.warning(
            codes='EDINET.EC5710W.FRTA.3.1.10',
            msg=_("Role types defined in the extension taxonomy must have a definition."),
            modelObject=errors,
        )


@validation(
    hook=ValidationHook.XBRL_FINALLY,
    disclosureSystems=[DISCLOSURE_SYSTEM_EDINET],
)
def rule_frta_4_2_2(
        pluginData: PluginValidationDataExtension,
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
) -> Iterable[Validation]:
    """
    EDINET.EC5710W: [FRTA.4.2.2] Taxonomy schemas must be defined in XML documents in which the XML Schema 'schema'
                                 element appears once only as the root element.
    """
    for modelDocument in val.modelXbrl.urlDocs.values():
        if pluginData.isStandardTaxonomyUrl(modelDocument.uri, val.modelXbrl):
            continue
        # check for nested Schema declarations which are not allowed.
        schemaElts = {elt for elt in modelDocument.xmlRootElement.iterdescendants(XbrlConst.qnXsdSchema.clarkNotation)}
        if len(schemaElts) > 0:
            yield Validation.warning(
                codes='EDINET.EC5710W.FRTA.4.2.2',
                msg=_("The root of a taxonomy schema file MUST be the XMLSchema element."),
                modelObject=modelDocument,
            )


@validation(
    hook=ValidationHook.XBRL_FINALLY,
    disclosureSystems=[DISCLOSURE_SYSTEM_EDINET],
)
def rule_frta_4_2_4(
        pluginData: PluginValidationDataExtension,
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
) -> Iterable[Validation]:
    """
    EDINET.EC5710W: [FRTA.4.2.4] Taxonomy schemas must declare elementFormDefault to be 'qualified',
                                 attributeFormDefault must have the value 'unqualified', and the 'form' attribute
                                  must not appear on element and attribute declarations.
    """
    for modelDocument in val.modelXbrl.urlDocs.values():
        if pluginData.isStandardTaxonomyUrl(modelDocument.uri, val.modelXbrl) or not modelDocument.type == ModelDocumentType.SCHEMA:
            continue
        rootElt = modelDocument.xmlRootElement
        if rootElt.get('elementFormDefault') != 'qualified' or rootElt.get('attributeFormDefault') != 'unqualified':
            yield Validation.warning(
                codes='EDINET.EC5710W.FRTA.4.2.4',
                msg=_("The XMLSchema root in taxonomy schema files must have the 'elementFormDefault' attribute set as "
                      "'qualified' and the 'attributeFormDefault' attribute set as 'unqualified'"),
                modelObject=modelDocument,
            )
        formUsages = []
        for elt in rootElt.iterdescendants([XbrlConst.qnXsdElement.clarkNotation, XbrlConst.qnXsdAttribute.clarkNotation]):
            if elt.get('form') is not None:
                formUsages.append(elt)
        if len(formUsages) > 0:
            yield Validation.warning(
                codes='EDINET.EC5710W.FRTA.4.2.4',
                msg=_("The 'form' attribute is not allowed on 'xsd:element' or 'xsd:attribute' declarations in a schema file"),
                modelObject=formUsages,
            )


@validation(
    hook=ValidationHook.XBRL_FINALLY,
    disclosureSystems=[DISCLOSURE_SYSTEM_EDINET],
)
def rule_frta_4_2_7(
        pluginData: PluginValidationDataExtension,
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
) -> Iterable[Validation]:
    """
    EDINET.EC5710W: [FRTA.4.2.7] A label linkbase should only contain labels defined in a single language.
    """
    for modelDocument in val.modelXbrl.urlDocs.values():
        if pluginData.isStandardTaxonomyUrl(modelDocument.uri, val.modelXbrl) or not modelDocument.type == ModelDocumentType.LINKBASE:
            continue
        usedLangs = {
            elt.get(XbrlConst.qnXmlLang.clarkNotation)
            for elt in modelDocument.xmlRootElement.iterdescendants(XbrlConst.qnLinkLabel.clarkNotation)
        }
        if len(usedLangs) > 1:
            yield Validation.warning(
                codes='EDINET.EC5710W.FRTA.4.2.7',
                msg=_("A label linkbase should only contain labels defined in a single language. This linkbase uses the following languages: %(langs)s"),
                langs=usedLangs,
                modelObject=modelDocument,
            )


@validation(
    hook=ValidationHook.XBRL_FINALLY,
    disclosureSystems=[DISCLOSURE_SYSTEM_EDINET],
)
def rule_frta_4_2_11(
        pluginData: PluginValidationDataExtension,
        val: ValidateXbrl,
        *args: Any,
        **kwargs: Any,
) -> Iterable[Validation]:
    """
    EDINET.EC5710W: [FRTA.4.2.11] Every schema in a DTS must define a non-empty targetNamespace attribute value
    """
    for modelDocument in val.modelXbrl.urlDocs.values():
        if pluginData.isStandardTaxonomyUrl(modelDocument.uri, val.modelXbrl) or not modelDocument.type == ModelDocumentType.SCHEMA:
            continue
        rootElt = modelDocument.xmlRootElement
        if rootElt.get('targetNamespace') is None or rootElt.get('targetNamespace') == "":
            yield Validation.warning(
                codes='EDINET.EC5710W.FRTA.4.2.11',
                msg=_("Every schema in a DTS must define a non-empty targetNamespace attribute value."),
                modelObject=modelDocument,
            )



"""
See COPYRIGHT.md for copyright information.
"""

from __future__ import annotations


class NamespaceConfig:
    def __init__(self) -> None:
        self.jpcrpEsr = "http://disclosure.edinet-fsa.go.jp/taxonomy/jpcrp-esr/2024-11-01/jpcrp-esr_cor"
        self.jpcrp = "http://disclosure.edinet-fsa.go.jp/taxonomy/jpcrp/2024-11-01/jpcrp_cor"
        self.jpcrpSbr = "http://disclosure.edinet-fsa.go.jp/taxonomy/jpcrp-sbr/2024-11-01/jpcrp-sbr_cor"
        self.jpctl = "http://disclosure.edinet-fsa.go.jp/taxonomy/jpctl/2024-11-01/jpctl_cor"
        self.jpdei = "http://disclosure.edinet-fsa.go.jp/taxonomy/jpdei/2013-08-31/jpdei_cor"
        self.jpigp = "http://disclosure.edinet-fsa.go.jp/taxonomy/jpigp/2024-11-01/jpigp_cor"
        self.jplvh = "http://disclosure.edinet-fsa.go.jp/taxonomy/jplvh/2024-11-01/jplvh_cor"
        self.jppfs = "http://disclosure.edinet-fsa.go.jp/taxonomy/jppfs/2024-11-01/jppfs_cor"
        self.jpspsEsr = "http://disclosure.edinet-fsa.go.jp/taxonomy/jpsps-esr/2024-11-01/jpsps-esr_cor"
        self.jpspsSbr = "http://disclosure.edinet-fsa.go.jp/taxonomy/jpsps-sbr/2024-11-01/jpsps-sbr_cor"
        self.jpsps = "http://disclosure.edinet-fsa.go.jp/taxonomy/jpsps/2024-11-01/jpsps_cor"
        self.jptoi = "http://disclosure.edinet-fsa.go.jp/taxonomy/jptoi/2024-11-01/jptoi_cor"
        self.jptooPst = "http://disclosure.edinet-fsa.go.jp/taxonomy/jptoo-pst/2024-11-01/jptoo-pst_cor"
        self.jptooToa = "http://disclosure.edinet-fsa.go.jp/taxonomy/jptoo-toa/2024-11-01/jptoo-toa_cor"
        self.jptooTon = "http://disclosure.edinet-fsa.go.jp/taxonomy/jptoo-ton/2024-11-01/jptoo-ton_cor"
        self.jptooTor = "http://disclosure.edinet-fsa.go.jp/taxonomy/jptoo-tor/2024-11-01/jptoo-tor_cor"
        self.jptooWto = "http://disclosure.edinet-fsa.go.jp/taxonomy/jptoo-wto/2024-11-01/jptoo-wto_cor"

        self._namespaceMap = {
            "jpcrp-esr_cor": self.jpcrpEsr,
            "jpcrp-sbr_cor": self.jpcrpSbr,
            "jpcrp_cor": self.jpcrp,
            "jpctl_cor": self.jpctl,
            "jpdei_cor": self.jpdei,
            "jpigp_cor": self.jpigp,
            "jplvh_cor": self.jplvh,
            "jppfs_cor": self.jppfs,
            "jpsps_cor": self.jpsps,
            "jpsps-esr_cor": self.jpspsEsr,
            "jpsps-sbr_cor": self.jpspsSbr,
            "jptoi_cor": self.jptoi,
            "jptoo-pst_cor": self.jptooPst,
            "jptoo-toa_cor": self.jptooToa,
            "jptoo-ton_cor": self.jptooTon,
            "jptoo-tor_cor": self.jptooTor,
            "jptoo-wto_cor": self.jptooWto,
        }

    def get(self, prefix: str) -> str | None:
        """get namespace URI for prefix"""
        return self._namespaceMap.get(prefix)



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from arelle.Cntlr import Cntlr
from arelle.ModelValue import qname
from arelle.ValidateXbrl import ValidateXbrl
from arelle.typing import TypeGetText
from arelle.utils.validate.ValidationPlugin import ValidationPlugin
from .PluginValidationDataExtension import PluginValidationDataExtension

import regex as re

_: TypeGetText

DANISH_CURRENCY_ID = 'DKK'
NAMESPACE_ARR = 'http://xbrl.dcca.dk/arr'
NAMESPACE_CMN = 'http://xbrl.dcca.dk/cmn'
NAMESPACE_FSA = 'http://xbrl.dcca.dk/fsa'
NAMESPACE_GSD = 'http://xbrl.dcca.dk/gsd'
NAMESPACE_SOB = 'http://xbrl.dcca.dk/sob'
PERSONNEL_EXPENSE_THRESHOLD = 200000
ROUNDING_MARGIN = 1000
REQUIRED_STATEMENT_OF_CHANGES_IN_EQUITY_FACTS = 3
REQUIRED_DISCLOSURE_OF_EQUITY_FACTS = 2


class ValidationPluginExtension(ValidationPlugin):
    def newPluginData(self, cntlr: Cntlr, validateXbrl: ValidateXbrl | None) -> PluginValidationDataExtension:
        return PluginValidationDataExtension(
            self.name,
            accountingPolicyConceptQns=frozenset([
                qname(f'{{{NAMESPACE_FSA}}}DescriptionMethodsOfRecognitionAndMeasurementBasisForCashFlowsStatement'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfAccountingPoliciesRelatedToDerivativeFinancialInstruments'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfAnyFailureOfApplyingEliminationBetweenConsolidatedEnterprises'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfChangeInInventoriesOfFinishedGoodsWorkInProgressAndGoodsForResale'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfChangesInCompositionOfEntityActivities'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfEffectOfChangeInAccountingEstimatesOnAssetsLiabilitiesEquityFinancialPositionAndResults'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfEffectOfChangeInRecognitionAndMeasurementBasisOfAssetsAndLiabilitiesAsResultOfErrors'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfEffectOfDerogationFromProvisionsOfDanishFinancialStatementsActOnAssetsLiabilitiesEquityFinancialPositionAndResults'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfGeneralMattersRelatedToRecognitionMeasurementAndChangesInAccountingPolicies'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfImpairmentOfFinancialAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfIncomeFromOtherInvestmentsAndReceivablesThatAreFixedAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfKeyFiguresAndFinancialRatios'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfAmortisationOfNoncurrentAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfAssetsMeantForSale'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfCurrentTaxReceivablesAndLiabilities'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfDividends'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfForeignCurrencies'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfHedgingRecognisedExpectedToReceiveAndAssumedAssetsAndLiabilities'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfImpairmentLossesAndDepreciation'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfIncomeExceedCostAndCostExceedsIncomeForTheFinancialYear'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfInvestments'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfInvestmentsAsCurrentAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfLeases'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfLiquidationAccount'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfPrepayments'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisForDistributions'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisForInvestmentsInJointVentures'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisForInvestmentsInSubsidiariesAndAssociates'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisIncludingBasesUsedForRevaluationsDepreciationAmortisationEstimatedResidualValueUsefulLifeWritedownsUpwardAndDownwardAdjustments'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfAdministrativeExpenses'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfAssetsAndLiabilities'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfCashAndCashEquivalents'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfContractWorkInProgress'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfCostOfProduction'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfCostOfSales'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfDeferredIncomeAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfDeferredIncomeLiabilities'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfDistributionCosts'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfEmployeeBenefitExpense'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfEquity'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfExpensesFromInvestmentsInGroupEnterprisesAndAssociates'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfExternalExpenses'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfFinanceExpenses'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfFinanceIncome'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfFinanceIncomeAndExpenses'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfGainsLossesFromCurrentValueAdjustmentsOfBiologicalAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfGainsLossesFromCurrentValueAdjustmentsOfInvestmentProperty'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfGainsLossesFromCurrentValueAdjustmentsOfOtherInvestmentAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfGrossProfitLoss'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfIncomeAndExpensesFromInvestmentsInGroupEnterprisesAndAssociates'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfIncomeAndExpensesFromInvestmentsInJointVentures'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfIncomeFromInvestmentsInGroupEnterprisesAndAssociates'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfIncomeStatementItems'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfIndirectCostsOfProductionRecognisedUnderAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfIntangibleAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfInventories'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfInvestmentProperty'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfInvestments'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfLeaseholdImprovements'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfLiabilitiesOtherThanProvisions'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfOtherInvestmentAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfOtherOperatingExpenses'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfOtherOperatingIncome'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfOtherOperatingIncomeAndExpenses'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfOtherProvisions'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfPropertyCosts'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfPropertyPlantAndEquipment'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfProvisions'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfReceivables'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfResearchAndDevelopmentExpendituresRecognisedAsExpenses'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfRevenue'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfSpecialItems'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfTaxExpenses'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisOfTaxPayablesAndDeferredTax'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisUsedAtEstablishment'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfRecognitionAndMeasurementBasisUsedInBusinessCombinations'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfStatingKeyFiguresAndFinancialRatiosIncludedInManagementReview'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfMethodsOfTranslationOfForeignCurrencies'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfOtherKeyFiguresAndFinancialRatios'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfOtherTaxExpenses'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfOwnWorkCapitalised'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfPublicGrants'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfRawMaterialsAndConsumablesUsed'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfRecognitionAndMeasurementBasisForDiscontinuedOperationsBalanceSheet'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfRemainingPositiveOrNegativeBalancesRelatedToAcquisitionAndUnitingOfInterestsAndMethodsUsedInConnectionWithTheirCalculation'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfWritedownsOfCurrentAssetsThatExceedNormalWritedowns'),
                qname(f'{{{NAMESPACE_FSA}}}DisclosureOfAccountingPolicies'),
                qname(f'{{{NAMESPACE_FSA}}}DisclosureOfTrueAndFairViewAndGoingConcern'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfAmortizationPeriodForGoodwill'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfAssumptionsOnWhichChosenCalculationMethodHasBeenBasedForFinancialAssetsAndLiabilitiesMeasuredAtNetPresentValueOrAmortisedCost'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfAssumptionsOnWhichChosenCalculationMethodHasBeenBasedForInvestmentsAndBiologicalAssetsMeasuredAtFairValue'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfBasisOnWhichEquityInvestmentsInSubsidiariesAndAssociatesHaveBeenRecognisedAtCost'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfChangeInAccountingEstimates'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfChangeInRecognitionAndMeasurementBasisOfAssetsAndLiabilitiesAsResultOfErrors'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfDerogationFromProvisionsOfDanishFinancialStatementsActOnAssetsLiabilitiesEquityFinancialPositionAndResults'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfDifferenceInBalanceSheetDatesBetweenParentAndConsolidatedSubsidiariesAndDescriptionOfImportantEventsOccurringAtThatTime'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfDifferenceInMethodsOfRecognitionAndMeasurementBasisBetweenParentAndConsolidatedGroupEnterprise'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfEntitysDefinitionOfCashAndCashEquivalents'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfGroundsOfNoncomparabilityOrRestatement'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfNotDisclosingInformationOnAGroupEntity'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfOtherMethodsOfRecognitionAndMeasurementBasisForAssetsInPreviousPeriod'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfWhetherInterestIsIncludedInCost'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnChangesAndEffectsOfChangesOnRecognitionAndMeasurementBasisResultingFromChangesInAccountingEstimatesOrErrors'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnConsolidatedResultsAndEquityIfSubsidiaryHasNotBeenPreviouslyIncludedInConsolidatedFinancialStatementsAndParentHoldsInvestmentsInSubsidiarySolelyWithViewToItsSubsequentTransfer'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnConsolidations'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnHedging'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnIncentivePrograms'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnIntragroupTransactions'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnLeasingContracts'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnMinorityInterests'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnNoncomparabilityOrRestatement'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnOmissionOfConsolidatedFinancialStatement'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnReportingClassOfEntity'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnSegments'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnSignificantUncertainties'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnSpecificPrerequisitesRegardingDevelopmentProjectsAndTaxAssets'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnSubsequentEvents'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnSubsidiariesHavingPresentedTheAnnualReportEtcWithReferenceToTheDanishFinancialStatementsAct78aConcerningTheExceptionForReportingClassCMediumsizeSubsidiariesWhichChoosesToPresentTheAnnualReportEtcAccordingToReportingClassB'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnThePrincipalModificationsAsResultOfTheLiquidation'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnTimingDifferencesInCustomerPaymentsEspeciallyForUtilities'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnTreatmentOfAdditionalPaymentsFromAndRepaymentsToRelatedParties'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnTrueAndFairViewIfProvisionsOfDanishFinancialStatementsActAreNotSufficient'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnUncertaintiesRelatingToGoingConcern'),
                qname(f'{{{NAMESPACE_FSA}}}SubsidiariesHasPresentedTheAnnualReportEtc.WithReferenceToTheDanishFinancialStatementsAct78aConcerningTheExceptionForReportingClassCMediumsizeSubsidiariesWhichChoosesToPresentTheAnnualReportEtcAccordingToReportingClassB'),
            ]),
            addressOfSubmittingEnterprisePostcodeAndTownQn=qname(f'{{{NAMESPACE_GSD}}}AddressOfSubmittingEnterprisePostcodeAndTown'),
            addressOfSubmittingEnterpriseStreetAndNumberQn=qname(f'{{{NAMESPACE_GSD}}}AddressOfSubmittingEnterpriseStreetAndNumber'),
            allReportingPeriodsMemberQn=qname(f'{{{NAMESPACE_GSD}}}AllReportingPeriodsMember'),
            annualReportTypes=frozenset([
                'Årsrapport',
                'årsrapport',
                'Annual report',
                'Likvidationsregnskab',
                'likvidationsregnskab',
                'Liquidation accounts',
            ]),
            assetsQn=qname(f'{{{NAMESPACE_FSA}}}Assets'),
            auditedAssuranceReportsDanish='Andre erklæringer med sikkerhed',
            auditedAssuranceReportsEnglish='The independent auditor\'s reports (Other assurance Reports)',
            auditedExtendedReviewDanish='Erklæring om udvidet gennemgang',
            auditedExtendedReviewEnglish='Auditor\'s report on extended review',
            auditedFinancialStatementsDanish='Revisionspåtegning',
            auditedFinancialStatementsEnglish='Auditor\'s report on audited financial statements',
            auditedNonAssuranceReportsDanish='Andre erklæringer uden sikkerhed',
            auditedNonAssuranceReportsEnglish='Auditor\'s reports (Other non-assurance reports)',
            averageNumberOfEmployeesQn=qname(f'{{{NAMESPACE_FSA}}}AverageNumberOfEmployees'),
            balanceSheetQnLessThanOrEqualToAssets=frozenset([
                qname(f'{{{NAMESPACE_FSA}}}NoncurrentAssets'),
                qname(f'{{{NAMESPACE_FSA}}}IntangibleAssets'),
                qname(f'{{{NAMESPACE_FSA}}}CompletedDevelopmentProjects'),
                qname(f'{{{NAMESPACE_FSA}}}ConcessionsOriginatingFromDevelopmentProjects'),
                qname(f'{{{NAMESPACE_FSA}}}PatentsOriginatingFromDevelopmentProjects'),
                qname(f'{{{NAMESPACE_FSA}}}TrademarksOriginatingFromDevelopmentProjects'),
                qname(f'{{{NAMESPACE_FSA}}}OtherSimilarRightsOriginatingFromDevelopmentProjects'),
                qname(f'{{{NAMESPACE_FSA}}}AcquiredIntangibleAssets'),
                qname(f'{{{NAMESPACE_FSA}}}AcquiredConcessions'),
                qname(f'{{{NAMESPACE_FSA}}}AcquiredPatents'),
                qname(f'{{{NAMESPACE_FSA}}}AcquiredLicences'),
                qname(f'{{{NAMESPACE_FSA}}}AcquiredTrademarks'),
                qname(f'{{{NAMESPACE_FSA}}}AcquiredOtherSimilarRights'),
                qname(f'{{{NAMESPACE_FSA}}}Goodwill'),
                qname(f'{{{NAMESPACE_FSA}}}DevelopmentProjectsInProgressAndPrepaymentsForIntangibleAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DevelopmentProjectsInProgress'),
                qname(f'{{{NAMESPACE_FSA}}}PrepaymentsForIntangibleAssets'),
                qname(f'{{{NAMESPACE_FSA}}}PropertyPlantAndEquipment'),
                qname(f'{{{NAMESPACE_FSA}}}LandAndBuildings'),
                qname(f'{{{NAMESPACE_FSA}}}Land'),
                qname(f'{{{NAMESPACE_FSA}}}Buildings'),
                qname(f'{{{NAMESPACE_FSA}}}InvestmentProperty'),
                qname(f'{{{NAMESPACE_FSA}}}OtherInvestmentAssets'),
                qname(f'{{{NAMESPACE_FSA}}}PlantAndMachinery'),
                qname(f'{{{NAMESPACE_FSA}}}FixturesFittingsToolsAndEquipment'),
                qname(f'{{{NAMESPACE_FSA}}}BiologicalAssets'),
                qname(f'{{{NAMESPACE_FSA}}}LeaseholdImprovements'),
                qname(f'{{{NAMESPACE_FSA}}}Ships'),
                qname(f'{{{NAMESPACE_FSA}}}Planes'),
                qname(f'{{{NAMESPACE_FSA}}}RightofuseAssets'),
                qname(f'{{{NAMESPACE_FSA}}}PropertyPlantAndEquipmentInProgressAndPrepaymentsForPropertyPlantAndEquipment'),
                qname(f'{{{NAMESPACE_FSA}}}PropertyPlantAndEquipmentInProgress'),
                qname(f'{{{NAMESPACE_FSA}}}PrepaymentsForPropertyPlantAndEquipment'),
                qname(f'{{{NAMESPACE_FSA}}}LongtermInvestmentsAndReceivables'),
                qname(f'{{{NAMESPACE_FSA}}}LongtermInvestmentsInGroupEnterprises'),
                qname(f'{{{NAMESPACE_FSA}}}LongtermInvestmentsInAssociates'),
                qname(f'{{{NAMESPACE_FSA}}}LongtermParticipatingInterests'),
                qname(f'{{{NAMESPACE_FSA}}}LongtermInvestmentsInJointVentures'),
                qname(f'{{{NAMESPACE_FSA}}}LongtermReceivablesFromGroupEnterprises'),
                qname(f'{{{NAMESPACE_FSA}}}LongtermReceivablesFromAssociates'),
                qname(f'{{{NAMESPACE_FSA}}}LongtermReceivablesFromParticipatingInterests'),
                qname(f'{{{NAMESPACE_FSA}}}LongtermReceivablesFromJointVentures'),
                qname(f'{{{NAMESPACE_FSA}}}OtherLongtermInvestments'),
                qname(f'{{{NAMESPACE_FSA}}}OtherLongtermReceivables'),
                qname(f'{{{NAMESPACE_FSA}}}LongtermReceivablesFromOwnersAndManagement'),
                qname(f'{{{NAMESPACE_FSA}}}NoncurrentDeferredTaxAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DepositsLongtermInvestmentsAndReceivables'),
                qname(f'{{{NAMESPACE_FSA}}}CostExceedsIncomeForTheFinancialYearLongtermReceivables'),
                qname(f'{{{NAMESPACE_FSA}}}ContributedCapitalInArrearsLongTerm'),
                qname(f'{{{NAMESPACE_FSA}}}NoncurrentContractAssets'),
                qname(f'{{{NAMESPACE_FSA}}}CurrentAssets'),
                qname(f'{{{NAMESPACE_FSA}}}Inventories'),
                qname(f'{{{NAMESPACE_FSA}}}RawMaterialsAndConsumables'),
                qname(f'{{{NAMESPACE_FSA}}}WorkInProgress'),
                qname(f'{{{NAMESPACE_FSA}}}ManufacturedGoodsAndGoodsForResale'),
                qname(f'{{{NAMESPACE_FSA}}}PrepaymentsForGoods'),
                qname(f'{{{NAMESPACE_FSA}}}Livestock'),
                qname(f'{{{NAMESPACE_FSA}}}PropertyHeldForSaleInTheOrdinaryCourseOfBusiness'),
                qname(f'{{{NAMESPACE_FSA}}}AssetsHeldForSaleInventories'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermReceivables'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermTradeReceivables'),
                qname(f'{{{NAMESPACE_FSA}}}ContractWorkInProgress'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermReceivablesFromGroupEnterprises'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermReceivablesFromAssociates'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermReceivablesFromJointVentures'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermReceivablesFromParticipatingInterests'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermReceivablesDividendsFromGroupEnterprises'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermReceivablesDividendsFromAssociates'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermReceivablesDividendsFromJointVentures'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermReceivablesDividendsFromParticipatingInterests'),
                qname(f'{{{NAMESPACE_FSA}}}CurrentDeferredTaxAssets'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermTaxReceivables'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermTaxReceivablesFromGroupEnterprises'),
                qname(f'{{{NAMESPACE_FSA}}}VatAndDutiesReceivables'),
                qname(f'{{{NAMESPACE_FSA}}}OtherShorttermReceivables'),
                qname(f'{{{NAMESPACE_FSA}}}ContributedCapitalInArrears'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermReceivablesFromOwnersAndManagement'),
                qname(f'{{{NAMESPACE_FSA}}}DeferredIncomeAssets'),
                qname(f'{{{NAMESPACE_FSA}}}CostExceedsIncomeForTheFinancialYearShorttermReceivables'),
                qname(f'{{{NAMESPACE_FSA}}}TimingDifferencesShorttermReceivablesEspeciallyUtilities'),
                qname(f'{{{NAMESPACE_FSA}}}CurrentContractAssets'),
                qname(f'{{{NAMESPACE_FSA}}}DerivativeFinancialInstrumentsShorttermAssets'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermInvestments'),
                qname(f'{{{NAMESPACE_FSA}}}ShorttermInvestmentsInGroupEnterprises'),
                qname(f'{{{NAMESPACE_FSA}}}OtherShorttermInvestments'),
                qname(f'{{{NAMESPACE_FSA}}}CashAndCashEquivalents'),
                qname(f'{{{NAMESPACE_FSA}}}AssetsMeantForSale'),
            ]),
            basisForAdverseOpinionDanish='Grundlag for afkræftende konklusion',
            basisForAdverseOpinionEnglish='Basis for Adverse Opinion',
            basisForDisclaimerOpinionDanish='Grundlag for manglende konklusion',
            basisForDisclaimerOpinionEnglish='Basis for Disclaimer of Opinion',
            basisForQualifiedOpinionDanish='Grundlag for konklusion med forbehold',
            basisForQualifiedOpinionEnglish='Basis for Qualified Opinion',
            classOfReportingEntityQn=qname(f'{{{NAMESPACE_FSA}}}ClassOfReportingEntity'),
            cClassOfReportingEntityEnums = frozenset([
                'SelectedElementsFromReportingClassD',
                'Reporting class C, medium-size enterprise',
                'Regnskabsklasse C, stor virksomhed',
                'Reporting class C, large enterprise']),
            dClassOfReportingEntityEnums = frozenset([
                'Regnskabsklasse D',
                'Reporting class D'
            ]),
            consolidatedMemberQn=qname(f'{{{NAMESPACE_CMN}}}ConsolidatedMember'),
            consolidatedSoloDimensionQn=qname(f'{{{NAMESPACE_CMN}}}ConsolidatedSoloDimension'),
            cpr_regex=re.compile(r'^([0-2][0-9]|3[0-1])(0[1-9]|1[0-2])[0-9]{2}-[0-9]{4}'),
            dateOfApprovalOfAnnualReportQn=qname(f'{{{NAMESPACE_SOB}}}DateOfApprovalOfAnnualReport'),
            dateOfExtraordinaryDividendDistributedAfterEndOfReportingPeriod=qname(f'{{{NAMESPACE_FSA}}}DateOfExtraordinaryDividendDistributedAfterEndOfReportingPeriod'),
            dateOfGeneralMeetingQn=qname(f'{{{NAMESPACE_GSD}}}DateOfGeneralMeeting'),
            declarationObligationQns=frozenset([
                qname(f'{{{NAMESPACE_ARR}}}ReportingResponsibilitiesAccordingToTheDanishExecutiveOrderOnApprovedAuditorsReportsEspeciallyTheCriminalCodeAndFiscalTaxAndSubsidyLegislationAudit'),
                qname(f'{{{NAMESPACE_ARR}}}ReportingResponsibilitiesAccordingToTheDanishExecutiveOrderOnApprovedAuditorsReportsEspeciallyTheCompaniesActOrEquivalentLegislationThatTheCompanyIsSubjectToAudit'),
                qname(f'{{{NAMESPACE_ARR}}}ReportingResponsibilitiesAccordingToTheDanishExecutiveOrderOnApprovedAuditorsReportsEspeciallyLegislationOnFinancialReportingIncludingAccountingAndStorageOfAccountingRecordsAudit'),
                qname(f'{{{NAMESPACE_ARR}}}ReportingResponsibilitiesAccordingToTheDanishExecutiveOrderOnApprovedAuditorsReportsEspeciallyOtherMattersAudit'),
                ]),
            descriptionOfQualificationsOfAssuranceEngagementPerformedQn=qname(f'{{{NAMESPACE_ARR}}}DescriptionOfQualificationsOfAssuranceEngagementPerformed'),
            descriptionOfQualificationsOfAuditedFinancialStatementsQn=qname(f'{{{NAMESPACE_ARR}}}DescriptionOfQualificationsOfAuditedFinancialStatements'),
            descriptionOfQualificationsOfFinancialStatementsExtendedReviewQn=qname(f'{{{NAMESPACE_ARR}}}DescriptionOfQualificationsOfFinancialStatementsExtendedReview'),
            descriptionsOfQualificationsOfReviewedFinancialStatementsQn=qname(f'{{{NAMESPACE_ARR}}}DescriptionsOfQualificationsOfReviewedFinancialStatements'),
            disclosureOfEquityQns=frozenset([
                qname(f'{{{NAMESPACE_FSA}}}InformationOnClassesOfIssuedShares'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfClassOfIssuedShares'),
                qname(f'{{{NAMESPACE_FSA}}}NumberOfIssuedShares'),
                qname(f'{{{NAMESPACE_FSA}}}NominelValuePerUnitShare'),
                qname(f'{{{NAMESPACE_FSA}}}AccountingParValuePerUnitShare'),
                qname(f'{{{NAMESPACE_FSA}}}NominalValueOfIssuedShares'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnExpensesIncurredInFormationOfEntityOrIncreaseInShareCapital'),
                qname(f'{{{NAMESPACE_FSA}}}ExpensesIncurredInFormationOfEntityOrIncreaseInShareCapital'),
                qname(f'{{{NAMESPACE_FSA}}}RetrospectiveInformationOnContributedCapital'),
                qname(f'{{{NAMESPACE_FSA}}}IncreaseDecreaseOfCapital'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfTreasurySharesHold'),
                qname(f'{{{NAMESPACE_FSA}}}NumberOfTreasurySharesHold'),
                qname(f'{{{NAMESPACE_FSA}}}NominalValueOfTreasurySharesHold'),
                qname(f'{{{NAMESPACE_FSA}}}AccountingParValueOfTreasurySharesHold'),
                qname(f'{{{NAMESPACE_FSA}}}PercentageOfContributedCapitalOfTreasurySharesHold'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnTreasurySharesAcquired'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfTreasurySharesAcquired'),
                qname(f'{{{NAMESPACE_FSA}}}NumberOfTreasurySharesAcquired'),
                qname(f'{{{NAMESPACE_FSA}}}NominalValueOfTreasurySharesAcquired'),
                qname(f'{{{NAMESPACE_FSA}}}AccountingParValueOfTreasurySharesAcquired'),
                qname(f'{{{NAMESPACE_FSA}}}PercentageOfContributedCapitalOfTreasurySharesAcquired'),
                qname(f'{{{NAMESPACE_FSA}}}PurchasePriceOfTreasurySharesAcquired'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfReasonForAcquisitionOfTreasuryShares'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnTreasurySharesSold'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfTreasurySharesSold'),
                qname(f'{{{NAMESPACE_FSA}}}NumberOfTreasurySharesSold'),
                qname(f'{{{NAMESPACE_FSA}}}NominalValueOfTreasurySharesSold'),
                qname(f'{{{NAMESPACE_FSA}}}PercentageOfContributedCapitalOfTreasurySharesSold'),
                qname(f'{{{NAMESPACE_FSA}}}SalesPriceOfTreasurySharesSold'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnTreasurySharesAcquiredAsSecurity'),
                qname(f'{{{NAMESPACE_FSA}}}DescriptionOfTreasurySharesAcquiredAsSecurity'),
                qname(f'{{{NAMESPACE_FSA}}}NumberOfTreasurySharesAcquiredAsSecurity'),
                qname(f'{{{NAMESPACE_FSA}}}NominalValueOfTreasurySharesAcquiredAsSecurity'),
                qname(f'{{{NAMESPACE_FSA}}}AccountingParValueOfTreasurySharesAcquiredAsSecurity'),
                qname(f'{{{NAMESPACE_FSA}}}PercentageOfContributedCapitalOfTreasurySharesAcquiredAsSecurity'),
                qname(f'{{{NAMESPACE_FSA}}}PurchasePriceOfTreasurySharesAcquiredAsSecurity'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfReasonForAcquisitionOfTreasurySharesAsSecurity'),
                qname(f'{{{NAMESPACE_FSA}}}ExplanationOfChangeInMinorityInterestsProportionateShareOfSubsidiariesEquity'),
                qname(f'{{{NAMESPACE_FSA}}}DisclosureOfDistributions'),
                qname(f'{{{NAMESPACE_FSA}}}Distributions'),
                qname(f'{{{NAMESPACE_FSA}}}DisclosureOfHedgeFund'),
                qname(f'{{{NAMESPACE_FSA}}}HedgeFund'),
                qname(f'{{{NAMESPACE_FSA}}}DisclosureOfReserveFund'),
                qname(f'{{{NAMESPACE_FSA}}}ReserveFund'),
                qname(f'{{{NAMESPACE_FSA}}}DisclosureOfReservesAvailable'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromReservesAvailable'),
                qname(f'{{{NAMESPACE_FSA}}}DisclosureOfLiquidationAccount'),
                qname(f'{{{NAMESPACE_FSA}}}LiquidationAccount'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnReserveForEntrepreneurialCompany'),
                qname(f'{{{NAMESPACE_FSA}}}ReserveForEntrepreneurialCompany'),
                qname(f'{{{NAMESPACE_FSA}}}InformationOnChangesAsResultOfApplicationOfUnitingOfInterestsMethod'),
                qname(f'{{{NAMESPACE_FSA}}}IncreaseDecreaseOfReservesThroughApplicationOfUnitingOfInterestsMethodEquity'),
                qname(f'{{{NAMESPACE_FSA}}}DisclosureOfRevaluationReserve'),
                qname(f'{{{NAMESPACE_FSA}}}RevaluationReserve'),
                qname(f'{{{NAMESPACE_FSA}}}AdditionsToRevaluationReserveInReportingPeriod'),
                qname(f'{{{NAMESPACE_FSA}}}ReversalsOfPriorYearsRevaluationsOfRevaluationReserve'),
                qname(f'{{{NAMESPACE_FSA}}}AmortisationOfRevaluationReserveInReportingPeriod'),
                qname(f'{{{NAMESPACE_FSA}}}DisposalsOfRevaluationReserveInReportingPeriod'),
                qname(f'{{{NAMESPACE_FSA}}}RecognitionAndMeasurementOfTheTaxEffectOfTheMovementsInTheRevaluationReserve'),
                qname(f'{{{NAMESPACE_FSA}}}CurrentValueReserve'),
                qname(f'{{{NAMESPACE_FSA}}}AdjustmentsOfCurrentValueReserveGainsLoss'),
                qname(f'{{{NAMESPACE_FSA}}}CurrentValueReserveRealized'),
                qname(f'{{{NAMESPACE_FSA}}}CurrentValueReserveReversed'),
            ]),
                distributionOfResultsQns=frozenset([
                qname(f'{{{NAMESPACE_FSA}}}DistributionsResultDistribution'),
                qname(f'{{{NAMESPACE_FSA}}}ExtraordinaryDistributions'),
                qname(f'{{{NAMESPACE_FSA}}}ProposedDividendRecognisedInEquity'),
                qname(f'{{{NAMESPACE_FSA}}}ProposedExtraordinaryDividendRecognisedInEquity'),
                qname(f'{{{NAMESPACE_FSA}}}ProposedExtraordinaryDividendRecognisedInLiabilities'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredFromToHedgeFund'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredFromToReserveFund'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredFromToReservesAvailable'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromEquityAttributableToParent'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromMinorityInterests'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromOtherStatutoryReserves'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromReserveAccordingToArticlesOfAssociation'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromReserveForNetRevaluationAccordingToEquityMethod'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromReserveForNetRevaluationOfInvestmentAssets'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromRestOfOtherReserves'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToFromRetainedEarnings'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToReserveForCurrentValueAdjustmentsOfCurrencyGains'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToReserveForCurrentValueOfHedging'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToReserveForDevelopmentExpenditure'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredToReserveForEntrepreneurialCompany'),
            ]),
            endDateForUseOfDigitalNonregisteredBookkeepingSystemQn=qname(f'{{{NAMESPACE_GSD}}}EndDateForUseOfDigitalNonregisteredBookkeepingSystem'),
            endDateForUseOfDigitalStandardBookkeepingSystemQn=qname(f'{{{NAMESPACE_GSD}}}EndDateForUseOfDigitalStandardBookkeepingSystem'),
            employeeBenefitsExpenseQn=qname(f'{{{NAMESPACE_FSA}}}EmployeeBenefitsExpense'),
            equityQn=qname(f'{{{NAMESPACE_FSA}}}Equity'),
            extraordinaryCostsQn=qname(f'{{{NAMESPACE_FSA}}}ExtraordinaryCosts'),
            extraordinaryIncomeQn=qname(f'{{{NAMESPACE_FSA}}}ExtraordinaryIncome'),
            extraordinaryResultBeforeTaxQn=qname(f'{{{NAMESPACE_FSA}}}ExtraordinaryResultBeforeTax'),
            fr37RestrictedText='has not given rise to reservations',
            forbiddenTypeOfSubmittedReportEnumerations=frozenset([
                'Selskabsselvangivelse',
                'Erklæring om undtagelse fra aflæggelse årsrapport',
                'ESG-rapport',
                'ESG report',
                ]),
            hasNotGivenRiseToReservationsText=frozenset([
                'har ikke givet anledning til forbehold',
                'has not given rise to reservations',
            ]),
            identificationNumberCvrOfAuditFirmQn=qname(f'{{{NAMESPACE_CMN}}}IdentificationNumberCvrOfAuditFirm'),
            identificationNumberCvrOfReportingEntityQn=qname(f'{{{NAMESPACE_GSD}}}IdentificationNumberCvrOfReportingEntity'),
            independentAuditorsReportDanish='Den uafhængige revisors erklæring',
            independentAuditorsReportEnglish='Independent Auditor’s Report',
            independentAuditorsReportReviewDanish='Den uafhængige revisors erklæringer (review)',
            independentAuditorsReportReviewEnglish='The independent auditor’s reports (Review)',
            independentPractitionersExtendedReviewReportDanish='Den uafhængige revisors erklæring om udvidet gennemgang',
            independentPractitionersExtendedReviewReportEnglish='Independent Practitioner’s Extended Review Report',
            independentPractitionersReviewReportDanish='Den uafhængige revisors reviewerklæring',
            independentPractitionersReviewReportEnglish='Independent Practitioner’s Review Report',
            informationOnTypeOfSubmittedReportQn=qname(f'{{{NAMESPACE_GSD}}}InformationOnTypeOfSubmittedReport'),
            legalEntityIdentifierOfReportingEntityQn=qname(f'{{{NAMESPACE_GSD}}}LegalEntityIdentifierOfReportingEntity'),
            liabilitiesQn=qname(f'{{{NAMESPACE_FSA}}}LiabilitiesAndEquity'),
            liabilitiesAndEquityQn=qname(f'{{{NAMESPACE_FSA}}}LiabilitiesAndEquity'),
            liabilitiesOtherThanProvisionsQn=qname(f'{{{NAMESPACE_FSA}}}LiabilitiesOtherThanProvisions'),
            longtermLiabilitiesOtherThanProvisionsQn=qname(f'{{{NAMESPACE_FSA}}}LongtermLiabilitiesOtherThanProvisions'),
            managementEndorsementQns=frozenset([
                qname(f'{{{NAMESPACE_SOB}}}StatementByExecutiveAndSupervisoryBoards'),
                qname(f'{{{NAMESPACE_SOB}}}TheReportingEntityAppliesTheExceptionConcerningOptingOutOfTheStatementByManagementEtc'),
                qname(f'{{{NAMESPACE_SOB}}}IdentificationOfApprovedAnnualReport'),
                qname(f'{{{NAMESPACE_SOB}}}ConfirmationThatAnnualReportIsPresentedInAccordanceWithRequirementsProvidedForByLegislationAnyStandardsAndRequirementsProvidedByArticlesOfAssociationOrByAgreement'),
                qname(f'{{{NAMESPACE_SOB}}}ConfirmationThatFinancialStatementGivesTrueAndFairViewOfAssetsLiabilitiesEquityFinancialPositionAndResults'),
                qname(f'{{{NAMESPACE_SOB}}}ConfirmationThatSupplementaryReportsGiveTrueAndFairViewInAccordanceWithGenerallyAcceptedGuidelinesForSuchReports'),
                qname(f'{{{NAMESPACE_SOB}}}ManagementsStatementAboutManagementsReview'),
                qname(f'{{{NAMESPACE_SOB}}}RecommendationForApprovalOfAnnualReportByGeneralMeeting'),
                qname(f'{{{NAMESPACE_SOB}}}DateOfApprovalOfAnnualReport'),
                qname(f'{{{NAMESPACE_SOB}}}PlaceOfSignatureOfStatement'),
                qname(f'{{{NAMESPACE_CMN}}}NameAndSurnameOfMemberOfExecutiveBoard'),
                qname(f'{{{NAMESPACE_CMN}}}TitleOfMemberOfExecutiveBoard'),
                qname(f'{{{NAMESPACE_CMN}}}DescriptionOfMemberOfExecutiveBoard'),
                qname(f'{{{NAMESPACE_SOB}}}DescriptionOfObjectionsOfMemberOfExecutiveBoardOnAnnualReport'),
                qname(f'{{{NAMESPACE_SOB}}}DescriptionOfSpecificAndAdequateGroundsForObjectionsOfMemberOfExecutiveBoardOnAnnualReport'),
                qname(f'{{{NAMESPACE_CMN}}}NameAndSurnameOfMemberOfSupervisoryBoard'),
                qname(f'{{{NAMESPACE_CMN}}}TitleOfMemberOfSupervisoryBoard'),
                qname(f'{{{NAMESPACE_CMN}}}DescriptionOfMemberOfSupervisoryBoard'),
                qname(f'{{{NAMESPACE_SOB}}}DescriptionOfObjectionsOfMemberOfSupervisoryBoardOnAnnualReport'),
                qname(f'{{{NAMESPACE_SOB}}}DescriptionOfSpecificAndAdequateGroundsForObjectionsOfMemberOfSupervisoryBoardOnAnnualReport'),
                qname(f'{{{NAMESPACE_SOB}}}IdentificationNumberCVRofParticipantNotPartOfTheRegisteredManagementMemberOfSupervisoryBoard'),
                qname(f'{{{NAMESPACE_SOB}}}NameOfEntityParticipantNotPartOfTheRegisteredManagementMemberOfSupervisoryBoard'),
                qname(f'{{{NAMESPACE_SOB}}}DescriptionOfParticipantNotPartOfTheRegisteredManagementMemberOfSupervisoryBoard'),
                qname(f'{{{NAMESPACE_GSD}}}NameAndSurnameOfMemberOfSupervisoryCommittee'),
                qname(f'{{{NAMESPACE_GSD}}}TitleOfMemberOfSupervisoryCommittee'),
                qname(f'{{{NAMESPACE_GSD}}}DescriptionOfMemberOfSupervisoryCommittee'),
                qname(f'{{{NAMESPACE_GSD}}}NameAndSurnameOfMemberOfSupervisoryCommittee'),
                qname(f'{{{NAMESPACE_GSD}}}TitleOfMemberOfSupervisoryCommittee'),
                qname(f'{{{NAMESPACE_CMN}}}NameAndSurnameOfLiquidator'),
                qname(f'{{{NAMESPACE_CMN}}}TitleOfLiquidator'),
                qname(f'{{{NAMESPACE_CMN}}}DescriptionOfLikvidator'),
                qname(f'{{{NAMESPACE_SOB}}}DescriptionOfObjectionsOfLiquidatorOnAnnualReport'),
                qname(f'{{{NAMESPACE_SOB}}}DescriptionOfSpecificAndAdequateGroundsForObjectionsOfLiquidatorOnAnnualReport'),
            ]),
            noncurrentAssetsQn=qname(f'{{{NAMESPACE_FSA}}}NoncurrentAssets'),
            nameAndSurnameOfChairmanOfGeneralMeetingQn=qname(f'{{{NAMESPACE_GSD}}}NameAndSurnameOfChairmanOfGeneralMeeting'),
            nameOfAuditFirmQn=qname(f'{{{NAMESPACE_CMN}}}NameOfAuditFirm'),
            nameOfReportingEntityQn=qname(f'{{{NAMESPACE_GSD}}}NameOfReportingEntity'),
            nameOfSubmittingEnterpriseQn=qname(f'{{{NAMESPACE_GSD}}}NameOfSubmittingEnterprise'),
            otherEmployeeExpenseQn=qname(f'{{{NAMESPACE_FSA}}}OtherEmployeeExpense'),
            otherRenderingOfReportedValueMemberQn=qname(f'{{{NAMESPACE_CMN}}}OtherRenderingOfReportedValueMember'),
            positiveProfitThreshold=1000,
            postemploymentBenefitExpenseQn=qname(f'{{{NAMESPACE_FSA}}}PostemploymentBenefitExpense'),
            precedingReportingPeriodEndDateQn=qname(f'{{{NAMESPACE_GSD}}}PredingReportingPeriodEndDate'),  # Typo in taxonomy
            precedingReportingPeriodStartDateQn=qname(f'{{{NAMESPACE_GSD}}}PrecedingReportingPeriodStartDate'),
            profitLossQn=qname(f'{{{NAMESPACE_FSA}}}ProfitLoss'),
            proposedDividendRecognisedInEquityQn=qname(f'{{{NAMESPACE_FSA}}}ProposedDividendRecognisedInEquity'),
            proposedExtraordinaryDividendRecognisedInLiabilitiesQn=qname(f'{{{NAMESPACE_FSA}}}ProposedExtraordinaryDividendRecognisedInLiabilities'),
            provisionsQn=qname(f'{{{NAMESPACE_FSA}}}Provisions'),
            registrationNumberOfTheDigitalStandardBookkeepingSystemUsedQn=qname(f'{{{NAMESPACE_GSD}}}RegistrationNumberOfTheDigitalStandardBookkeepingSystemUsed'),
            registeredReportingPeriodDeviatingFromReportedReportingPeriodDueArbitraryDatesMemberQn=qname(f'{{{NAMESPACE_GSD}}}RegisteredReportingPeriodDeviatingFromReportedReportingPeriodDueArbitraryDatesMember'),
            reportedValueOtherRenderingOfReportedValueDimensionQn=qname(f'{{{NAMESPACE_CMN}}}ReportedValueOtherRenderingOfReportedValueDimension'),
            reportingClassCLargeDanish='Regnskabsklasse C, stor virksomhed',
            reportingClassCLargeEnglish='Reporting class C, large enterprise',
            reportingClassCLargeLowercaseDanish='regnskabsklasse C, stor virksomhed',
            reportingClassCLargeLowercaseEnglish='reporting class C, large enterprise',
            reportingClassCMediumDanish='Regnskabsklasse C, mellemstor virksomhed',
            reportingClassCMediumEnglish='Reporting class C, medium-size enterprise',
            reportingClassCMediumLowercaseDanish='regnskabsklasse C, mellemstor virksomhed',
            reportingClassCMediumLowercaseEnglish='reporting class C, medium-size enterprise',
            reportingClassDDanish='Regnskabsklasse D',
            reportingClassDEnglish='Reporting class D',
            reportingClassDLowercaseDanish='regnskabsklasse D',
            reportingClassDLowercaseEnglish='reporting class D',
            reportingPeriodEndDateQn=qname(f'{{{NAMESPACE_GSD}}}ReportingPeriodEndDate'),
            reportingPeriodStartDateQn=qname(f'{{{NAMESPACE_GSD}}}ReportingPeriodStartDate'),
            reportingResponsibilitiesOnApprovedAuditorsReportAuditQn=qname(f'{{{NAMESPACE_ARR}}}ReportingResponsibilitiesAccordingToTheDanishExecutiveOrderOnApprovedAuditorsReportsAudit'),
            reportingResponsibilitiesOnApprovedAuditorsReportsExtendedReviewQn=qname(f'{{{NAMESPACE_ARR}}}ReportingResponsibilitiesAccordingToTheDanishExecutiveOrderOnApprovedAuditorsReportsExtendedReview'),
            reportingObligationQns = frozenset([
                qname(f'{{{NAMESPACE_ARR}}}ReportingResponsibilitiesAccordingToTheDanishExecutiveOrderOnApprovedAuditorsReportsEspeciallyTheCriminalCodeAndFiscalTaxAndSubsidyLegislationExtendedReview'),
                qname(f'{{{NAMESPACE_ARR}}}ReportingResponsibilitiesAccordingToTheDanishExecutiveOrderOnApprovedAuditorsReportsEspeciallyTheCompaniesActOrEquivalentLegislationThatTheCompanyIsSubjectToExtendedReview'),
                qname(f'{{{NAMESPACE_ARR}}}ReportingResponsibilitiesAccordingToTheDanishExecutiveOrderOnApprovedAuditorsReportsEspeciallyLegislationOnFinancialReportingInApplication'),
            ]),
            schemaRefUri='http://archprod.service.eogs.dk/taxonomy/',
            selectedElementsFromReportingClassCQn=qname(f'{{{NAMESPACE_FSA}}}SelectedElementsFromReportingClassC'),
            selectedElementsFromReportingClassDQn=qname(f'{{{NAMESPACE_FSA}}}SelectedElementsFromReportingClassD'),
            shorttermLiabilitiesOtherThanProvisionsQn=qname(f'{{{NAMESPACE_FSA}}}ShorttermLiabilitiesOtherThanProvisions'),
            signatureOfAuditorsDateQn=qname(f'{{{NAMESPACE_ARR}}}SignatureOfAuditorsDate'),
            soloMemberQn=qname(f'{{{NAMESPACE_CMN}}}SoloMember'),
            startDateForUseOfDigitalNonregisteredBookkeepingSystemQn=qname(f'{{{NAMESPACE_GSD}}}StartDateForUseOfDigitalNonregisteredBookkeepingSystem'),
            startDateForUseOfDigitalStandardBookkeepingSystemQn=qname(f'{{{NAMESPACE_GSD}}}StartDateForUseOfDigitalStandardBookkeepingSystem'),
            statementOfChangesInEquityQns=frozenset([
                qname(f'{{{NAMESPACE_FSA}}}Equity'),
                qname(f'{{{NAMESPACE_FSA}}}IncreaseDecreaseOfEquityThroughCorrectionsOfErrors'),
                qname(f'{{{NAMESPACE_FSA}}}IncreaseDecreaseOfEquityThroughChangesInAccountingPolicies'),
                qname(f'{{{NAMESPACE_FSA}}}AdditionsToDisposalsOfEquityThroughMergersAndBusinessCombinations'),
                qname(f'{{{NAMESPACE_FSA}}}IncreaseDecreaseOfEquityByDivisionAndSaleOfOtherCompany'),
                qname(f'{{{NAMESPACE_FSA}}}IncreaseOfCapital'),
                qname(f'{{{NAMESPACE_FSA}}}IncreaseOfCapitalByConversionOfDebt'),
                qname(f'{{{NAMESPACE_FSA}}}DecreaseOfCapital'),
                qname(f'{{{NAMESPACE_FSA}}}DecreaseOfCapitalDueToDeficitCoverage'),
                qname(f'{{{NAMESPACE_FSA}}}BonusShares'),
                qname(f'{{{NAMESPACE_FSA}}}CostRelatedToIncreaseOfCapital'),
                qname(f'{{{NAMESPACE_FSA}}}PurchaseOfTreasuryShares'),
                qname(f'{{{NAMESPACE_FSA}}}SalesOfTreasuryShares'),
                qname(f'{{{NAMESPACE_FSA}}}PurchaseOfMinorityShares'),
                qname(f'{{{NAMESPACE_FSA}}}SaleOfMinorityShares'),
                qname(f'{{{NAMESPACE_FSA}}}Dividend'),
                qname(f'{{{NAMESPACE_FSA}}}DividendPaid'),
                qname(f'{{{NAMESPACE_FSA}}}DistributionsStatementOfChangesInEquity'),
                qname(f'{{{NAMESPACE_FSA}}}ExtraordinaryDistributionsStatementOfChangesInEquity'),
                qname(f'{{{NAMESPACE_FSA}}}IncreaseDecreaseOfInvestmentsThroughNetExchangeDifferencesEquity'),
                qname(f'{{{NAMESPACE_FSA}}}ValueAdjustmentsOfEquity'),
                qname(f'{{{NAMESPACE_FSA}}}AdjustmentsOfHedgingInstruments'),
                qname(f'{{{NAMESPACE_FSA}}}OtherAdjustmentsOfEquity'),
                qname(f'{{{NAMESPACE_FSA}}}EquityTransfersToReserves'),
                qname(f'{{{NAMESPACE_FSA}}}ProfitLoss'),
                qname(f'{{{NAMESPACE_FSA}}}ChangesInEquityOfTax'),
                qname(f'{{{NAMESPACE_FSA}}}ExtraordinaryDividendPaid'),
                qname(f'{{{NAMESPACE_FSA}}}DividendsFromTreasuryShares'),
                qname(f'{{{NAMESPACE_FSA}}}ReceivablesShareCapitalPaid'),
                qname(f'{{{NAMESPACE_FSA}}}Revaluations'),
                qname(f'{{{NAMESPACE_FSA}}}DepreciationOfEquity'),
                qname(f'{{{NAMESPACE_FSA}}}RevaluationsOfActuarialProfitLossForReportingPeriod'),
                qname(f'{{{NAMESPACE_FSA}}}ImpairmentLossesAndDepreciationOfActuarialGainsLossesStatementOfChangesInEquity'),
                qname(f'{{{NAMESPACE_FSA}}}DissolutionOfPreviousYearsRevaluations'),
                qname(f'{{{NAMESPACE_FSA}}}DissolutionOfFairValueReserve'),
                qname(f'{{{NAMESPACE_FSA}}}DissolutionOfReserves'),
                qname(f'{{{NAMESPACE_FSA}}}AdjustmentsOfHedgingInstrumentsAtFairValue'),
                qname(f'{{{NAMESPACE_FSA}}}AdjustmentsOfCollateralsBySaleOfInvestments'),
                qname(f'{{{NAMESPACE_FSA}}}TransferredFromSharePremium'),
                qname(f'{{{NAMESPACE_FSA}}}TransferToCoverageOfLosses'),
                qname(f'{{{NAMESPACE_FSA}}}ProfitLlossFromGroupEnterprises'),
                qname(f'{{{NAMESPACE_FSA}}}ProfitLossFromAssociates'),
                qname(f'{{{NAMESPACE_FSA}}}ProfitLossFromParticipatingInterests'),
                qname(f'{{{NAMESPACE_FSA}}}ReversedRevaluationsDuringReportingPeriod'),
                qname(f'{{{NAMESPACE_FSA}}}DistributedDividendsFromGroupEnterprises'),
                qname(f'{{{NAMESPACE_FSA}}}DistributedDividendsFromAssociates'),
                qname(f'{{{NAMESPACE_FSA}}}DistributedDividendsFromParticipatingInterests'),
                qname(f'{{{NAMESPACE_FSA}}}CashPaymentsConcerningFormationOfEntity'),
                qname(f'{{{NAMESPACE_FSA}}}ContributionFromGroup'),
                qname(f'{{{NAMESPACE_FSA}}} IncreaseDecreaseOfReservesThroughApplicationOfUnitingOfInterestsMethodEquity')
                ]),
            taxExpenseOnOrdinaryActivitiesQn=qname(f'{{{NAMESPACE_FSA}}}TaxExpenseOnOrdinaryActivities'),
            taxExpenseQn=qname(f'{{{NAMESPACE_FSA}}}TaxExpense'),
            typeOfAuditorAssistanceQn=qname(f'{{{NAMESPACE_CMN}}}TypeOfAuditorAssistance'),
            typeOfBasisForModifiedOpinionOnFinancialStatementsReviewQn=qname(f'{{{NAMESPACE_ARR}}}TypeOfBasisForModifiedOpinionOnFinancialStatementsReview'),
            typeOfDigitalNonregisteredBookkeepingSystemQn=qname(f'{{{NAMESPACE_GSD}}}TypeOfDigitalNonregisteredBookkeepingSystem'),
            typeOfReportingPeriodDimensionQn=qname(f'{{{NAMESPACE_GSD}}}TypeOfReportingPeriodDimension'),
            wagesAndSalariesQn=qname(f'{{{NAMESPACE_FSA}}}WagesAndSalaries'),
        )
